#####################################################################
# 
# ZTxnQueue.py - Zope Transactional Queue 
#
# $Id: ZTxnQueue.py,v 1.19 2003/11/27 13:33:45 kevin Exp $
#
# Copyright (C) 2003 Kevin Gill
#
# Derived from the Xron version 0.0.10
# Copyright (C) 2000 Loren Stafford
#  
# Xron is Derived from the ZScheduler product version 0.0.7 in
# accord with the terms of the ZScheduler's license.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License  
# along with this program; if not, write to the 
#     Free Software Foundation, Inc., 
#     59 Temple Place - Suite 330, 
#     Boston, MA  02111-1307, USA.
# 
#####################################################################

""" Zope Transactional Queue """

import string
from types import StringType, UnicodeType

import TQInterface, TQPollingThread
from TQUtility import UniqueId

from Interface import Base
from ZODB.PersistentList import PersistentList
from Persistence import Persistent
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import ClassSecurityInfo, getSecurityManager, Role
from OFS.Folder import Folder
from OFS.PropertyManager import PropertyManager
from Globals import InitializeClass
import Acquisition
from Shared.DC.ZRDB.TM import TM
from zExceptions.ExceptionFormatter import format_exception
import zLOG

# I want to track dequeues - Multiple threads access the queue
# they will be blocked, being forced to access it serially. However,
# each instance of a blocked class will have stale data.
# _TrackedDequeues is a centralised store, where a stale
# object can identify the point up to which items have been dequeued.
# 
# The Last dequeued object id is here.
# If that is in your version of the data, simply wind forward
# to it. 
#
# I use the physical path to the zope object as an index to this
# file. This, I believe is inefficient and I also believe that
# there must be some better id around for this. I cannot use
# 'id' as it may be duplicated in a different folder. I cannot
# generate a key, because the object could be copied through ZMI.
_TrackedDequeues={}


# Each queue has a mutext to prevent multiple threads from reading
# from that queue at the same time
import ThreadLock
_mutex = ThreadLock.allocate_lock()
_AllMutexes={}
def _AcquireMutex(path):
    _mutex.acquire()
    try:
        if not _AllMutexes.has_key(path):
            _AllMutexes[path] = ThreadLock.allocate_lock()
        m = _AllMutexes[path]
    finally:
        _mutex.release()
    m.acquire()

def _ReleaseMutex(path):
    _mutex.acquire()
    try:
        _AllMutexes[path].release()
    finally:
        _mutex.release()

## I am using this in the sub-class also
security = ClassSecurityInfo()

## True and False in Python 2.1
try:
    a=True
except:
    True=1
    False=0

class ZTxnQueueBase(TM, PersistentList, Folder, PropertyManager, Acquisition.Implicit, Role.RoleManager):
    """Cheap and dirty implementation. I use a persistent mapping to store
    the queue, and the keys are sequential to maintain the queue ordering."""

    def enqueue(self, Data):
        """Enqueues a piece of data. """
        return self._enqueue(Data)
    def dequeue(self):
        """Pops an item off the queue. """
        return self._dequeue()
    def items(self):
        """View the Contents of the Queue - For the Management Screens"""
        self._AmIManaged()
        rv = []
        for i in self:
            rv.append((i[0], i[1]))
        return rv
    def find(self, OID):
        """Returns the data for OID or None if not found"""
        for (storedOID, storedData, process) in self:
            if storedOID == OID:
                return storedData
        return None
    def block(self):
        """Marks the queue as blocked"""
        self.ProcessingState='B'
        self._p_changed = 1
    def unblock(self):
        """Marks the queue as active"""
        self.ProcessingState='A'
        self._p_changed = 1

    def __init__(self, id, title):
        """Initialise """
        self.id = id
        self.path = None
        self.title = title
        PersistentList.__init__(self)

        # These are the things to enable a queue proccessing function
        self.ProcessingState='A' # A=Active, B=Blocked

        # This is the batch size
        self.QueueProcessSize=10

        # This is the data to be merged - If you want to subclass the
        # ZTxnQueue, and add new keys, add them to this list.
        self.persistentKeys=['path', 'title', 'ProcessingState', 'QueueProcessSize']
        # Track the last dequeued OID - For resolving conflict
        self.LastDequeued=None

        # Keep track of Errors. These can be displayed on a separate
        # tab for the administrator.
        self.LastError=None

    def sleep(self, seconds):
        """I just put this in so I could test concurrency easily"""
        import time
        time.sleep(seconds)

    ### Consider everything beyond this private

    __ac_roles__ = ()
    __implements__ = (TQInterface.IZTxnQueue, TQInterface.IPolledObject)
    meta_type = 'ZTxnQueue'

    security.declareProtected('ZTxnQueue Modify', 'enqueue')
    security.declareProtected('ZTxnQueue Modify', 'dequeue')
    security.declarePublic('items')
    security.declarePublic('find')
    security.setPermissionDefault('ZTxnQueue Modify', ('Manager', 'Owner'))

    manage_options = ((
       {'label':'Settings', 'action':'manage_editForm', 'help': ('ZTxnQueue', 'AZTxnQueue.stx')},
       {'label':'Contents', 'action':'manage_main', 'help': ('ZTxnQueue', 'AZTxnQueue.stx')},
       {'label':'Queue', 'action':'manage_viewForm', 'help': ('ZTxnQueue', 'AZTxnQueue.stx')},
       {'label':'Exception', 'action':'manage_exceptionForm', 'help': ('ZTxnQueue', 'AZTxnQueue.stx')},)+ 
       Role.RoleManager.manage_options)
        
    manage_viewForm = PageTemplateFile('www/manage_View.zpt', globals())
    manage_editForm = PageTemplateFile('www/manage_EditForm.zpt', globals())
    manage_exceptionForm = PageTemplateFile('www/manage_ExceptionForm.zpt', globals())
    _properties = ({'id': 'title', 'type': 'string', 'mode': 'w'},
                   {'id': 'path', 'type': 'string', 'mode': 'r'},
                   {'id': 'ProcessingState', 'type': 'string', 'mode': 'w'},
                   {'id': 'QueueProcessSize', 'type': 'string', 'mode': 'w'},
    )

    security.declarePrivate('_enqueue')
    def _enqueue(self, Data):
        """Place an item into the queue"""
        transaction = get_transaction()
        if hasattr(self, '_v_gotdequeuelock') and self._v_gotdequeuelock == 'Y':
            ## This is a dequeueing transaction - raise an exception
            raise "ZTxnQueue.WriteError", "Attempt to enqueue to a queue in a dequeue transaction"
        if not hasattr(self, '_v_registered') or self._v_registered != 'Y':
            self._registered = 0  # Seems to be problems tracking this variable - may be persisted?
            self._register()
            transaction.register(self)
        self._v_enqueuetrans = 'Y'
        self._p_changed = 1
        OID = UniqueId()
        self.append((OID, Data, True))
        return OID

    security.declarePrivate('_dequeue')
    def _dequeue(self):
        """Dequeue an item"""
        if hasattr(self, '_v_enqueuetrans') and self._v_enqueuetrans == 'Y':
            ## This is a enqueueing transaction - raise an exception
            raise "ZTxnQueue.ReadError", "Attempt to dequeue from a queue in an enqueue transaction"
        if not hasattr(self, '_v_gotdequeuelock') or self._v_gotdequeuelock == 'N':
            _AcquireMutex(self.path)
            self._v_gotdequeuelock = 'Y'
            # Where this object is older than the database version,
            # it will have items at the start that are already
            # processed - wind past these.
            try:
                lastDequeued = _TrackedDequeues[self.path]
            except:
                lastDequeued = None
            if lastDequeued and (lastDequeued != self.LastDequeued):
                index = 0
                for (oid, data, process) in self:
                    index+=1
                    if lastDequeued == oid:
                       del self[:index] # Remove the items up to / including oid
                       self._p_changed = 1
                       break
                self.LastDequeued = lastDequeued
            if not hasattr(self, '_v_registered') or self._v_registered != 'Y':
                # Need to register both for TM and Persistence
                self._registered = 0  # seems to be problems tracking this variable - may be persisted?
                self._register()                   
                get_transaction().register(self)
        # Finally, down to business
        try:
            rv = self.pop(0)
        except:
            return None
        self.LastDequeued=rv[0]
        self._p_changed = 1
        return (rv[0], rv[1])

    def _recordException(self, info, OID=None, Data=None, PD=None):
        """An exception occured. Store information about the exception so
        that it can be displayed through the ZMI."""
        import time
        
        ErrorLog = ["<h3>Exception</h3>"]
        ErrorLog.append('<table><tr><td align="right"><b>Time:</b></td><td>%s</td></tr>' % time.ctime())
        ErrorLog.append('<table><tr><td align="right"><b>Path:</b></td><td>%s</td></tr>' % self.path)

        strtype = str(getattr(info[0], '__name__', info[0]))
        ErrorLog.append('<table><tr><td align="right"><b>Type:</b></td><td>%s</td></tr>' % strtype)

        try:
            strv = str(info[1])
        except:
            strv = '<unprintable %s object>' % str(type(info[1]).__name__)
        ErrorLog.append('<table><tr><td align="right"><b>Value:</b></td><td>%s</td></tr>' % strv)

        if OID:
            ErrorLog.append('<table><tr><td align="right"><b>OID:</b></td><td>%s</td></tr>' % str(OID))
        if Data:
            ErrorLog.append('<table><tr><td align="right"><b>Data:</b></td><td>%s</td></tr>' % str(Data))
        if PD:
            ErrorLog.append('<table><tr><td align="right"><b>Private Data:</b></td><td>%s</td></tr>' % str(PD))

        ErrorLog.append("</table><br><br>")
        if not isinstance(info[2], StringType) and not isinstance(info[2], UnicodeType):
            ErrorLog.append(string.join(format_exception(*info, **{'as_html': 0}), "<br>"))

        self.LastError=string.join(ErrorLog, "\n")

    def _p_resolveConflict(self, oldState, savedState, newState):
        """This module contains a variable which tracks the last
        dequeued item. If we assume that OIDs are sequential, we
        can say that the resulting queue from merging the newState
        and the savedState must be everything with an OID
        greater than the lastDequeued item. - Of course this is
        except in the case of an abort during dequeuing."""
        LogMsg = ["_p_resolveConflict for [%s]"%newState['path']]
        if len(oldState['data']) > 0:
            LogMsg.append("oldState %d [%s ... %s]" %
                (len(oldState['data']), oldState['data'][0][0], oldState['data'][-1][0]))
        else:
            LogMsg.append("oldState is empty") 
        if len(newState['data']) > 0:
            LogMsg.append("newState %d [%s ... %s]" % 
                (len(newState['data']), newState['data'][0][0], newState['data'][-1][0]))
        else:
            LogMsg.append("newState is empty") 
        if len(savedState['data']) > 0:
            LogMsg.append("savedState %d [%s ... %s]" % 
                (len(savedState['data']), savedState['data'][0][0], savedState['data'][-1][0]))
        else:
            LogMsg.append("savedState is empty") 
        LogMsg.append("savedState['LastDequeued'] =%s, newState['LastDequeued'] = %s" %
            (savedState['LastDequeued'], newState['LastDequeued']))

        mergedData = []
        ## First we want to put the data which is committed on the 
        ## database into the merged list. If we are dequeuing, and the
        ## items that we dequeued are on the stored version, then
        ## we want to removed them.
        if newState['LastDequeued'] == oldState['LastDequeued']:
            ## Then we are enqueuing. Simply create the new
            ## data from the saved data + anything we have added
            mergedData = savedState['data'] + newState['data'][len(oldState['data']):]
        else:
            ## We are dequeueing
            ## Find our LastDequeued item in the savedData, 
            ## and use everything after it.
            if newState['LastDequeued'] == savedState['LastDequeued']:
                ## Special Case, we have just wound forward to the savedData
                ## view but not dequeued anything from this object...
                mergedData = savedState['data']
            else:
                ## We actually dequeued something
                lastOID = newState['LastDequeued']
                for index in range(len(savedState['data'])):
                    if savedState['data'][index][0] == lastOID:
                        mergedData = savedState['data'][index+1:]
                        break
                savedState['LastDequeued'] = newState['LastDequeued']

        if len(mergedData) > 0:
            LogMsg.append("Result %d [%s ... %s], LastDequeued = %s" %
                (len(mergedData), mergedData[0][0], mergedData[-1][0], savedState['LastDequeued']))
        else:
            LogMsg.append("Result is empty, LastDequeued = %s" % savedState['LastDequeued'])
        savedState['data'] = mergedData

        ## Thats data done, what about the other properties
        for k in savedState['persistentKeys']:
            if oldState.has_key(k):
                if oldState[k] != newState[k]:
                    savedState[k] = newState[k]
            else:
                savedState[k] = newState[k]
        zLOG.LOG('ZTxnQueue', zLOG.WARNING, string.join(LogMsg, '\n'))
        return savedState

    # I am using a MUTEX to ensure that only one
    # threads can get dequeue access to the object at a time.
    # If two threads started dequeuing, they would only 
    # find out after it was too late - i.e. on a commit() !! 
    # The _registered flag seems to get pickled under certain
    # situations, leading to transactions being disabled. I 
    # have a shadow variable _v_registered.
    def _begin(self):
        """Begin"""
        self._v_registered = 'Y'
    def _finish(self):
        """Commit"""
        if hasattr(self, "_v_gotdequeuelock") and self._v_gotdequeuelock == 'Y':
            _TrackedDequeues[self.path] = self.LastDequeued
        try:
            # There is some bizare behaviour in the case of an abort(). It seems to 
            # be called using an object which has lost _v_ variables
            _ReleaseMutex(self.path)
        except:
            pass
        self._v_gotdequeuelock = 'N'
        self._v_enqueuetrans = 'N'
        self._v_registered = 'N'
    def _abort(self):
        """Rollback"""
        if hasattr(self, "_v_gotdequeuelock") and self._v_gotdequeuelock == 'Y':
            # Reset the last dequeued OID to the initial state
            try:
                self.LastDequeued = _TrackedDequeues[self.path]
            except:
                self.LastDequeued = None
        try:
            # There is some bizare behaviour in the case of an abort(). It seems to 
            # be called using an object which has lost _v_ variables
            _ReleaseMutex(self.path)
        except:
            pass
        self._v_gotdequeuelock = 'N'
        self._v_enqueuetrans = 'N'
        self._v_registered = 'N'

    def _AmIManaged(self):
        """Verify that the current instance is managed. Also handle renames"""
        ManagerObj = getattr(self, TQPollingThread.GlobalID, None)
        if not ManagerObj: return
        current_path = string.join(self.getPhysicalPath(), "/")
        if not self.path:
            self.path = current_path
            self._p_changed = 1
            ManagerObj.registerObject(self)
        elif self.path != current_path:
            old_path = self.path
            self.path = current_path
            self._p_changed = 1
            ManagerObj.reregisterObject(old_path, self)
        else:
            ManagerObj.registerObject(self)

    def manage_beforeDelete(self, item, container):
        """Called when the queue is deleted - remove Polling Thread entry"""
        if not self.path:
            return
        ManagerObj = getattr(self, TQPollingThread.GlobalID, None)
        if not ManagerObj: return
        ManagerObj.unregisterObject(self)

    security.declareProtected('ZTxnQueue Modify', 'manage_clearQueue')
    def manage_clearQueue(self, REQUEST="None"):
        """Called to delete all entries from the queue. Use with care."""
        count = 0
        while 1:
            rv = self.dequeue()
            if not rv:
                break
            count = count + 1
        if REQUEST:
            REQUEST.set('manage_tabs_message', "%d Items have been deleted from the queue" % count)
            return self.manage_editForm(self,REQUEST)

    def manage_edit(self, title, REQUEST="None"):
        """Replacement for the default property management screen"""
        self._AmIManaged()
        self.title = title
        self._p_changed = 1
        if REQUEST:
            if REQUEST.has_key('ProcessingState'):
                self.ProcessingState = 'B'
            else:
                self.ProcessingState = 'A'
            try:
                QueueProcessSize = int(REQUEST['QueueProcessSize'].strip())
            except:
                REQUEST.set('manage_tabs_message', "Error: Queue Process Size must be an Integer ")
                return self.manage_editForm(self,REQUEST)

            if QueueProcessSize <= 0:
                REQUEST.set('manage_tabs_message', "Error: Queue Process Size must be greater than zero")
                return self.manage_editForm(self,REQUEST)
            self.QueueProcessSize = QueueProcessSize
            REQUEST.set('manage_tabs_message', "Configuration Changed.")
            return self.manage_editForm(self,REQUEST)

    def Process(self, App, private_data):
        """This method is called every 10 seconds (configurable) to
        process the queue stored in this object."""

        # Verify that there is work to do
        if self.ProcessingState != 'A': return 0
        if len(self.data) == 0: return 0
        if not hasattr(self, 'ProcessOne'):   # No processing object defined
            return 0

        try:
            QueueProcessSize = int(self.QueueProcessSize)
        except:
            zLOG.LOG('ZTxnQueue', zLOG.WARNING,
                "ZTxnQueue %s - Has invalid QueueProcessSize %s - Setting to 10" %
                (self.path, self.QueueProcessSize))
            get_transaction().abort()
            get_transaction().begin()
            QueueProcessSize = 10
            self.QueueProcessSize = 10
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()
            # Non-Fatal Problem - Exit process, but should be fine the next time through. 
            return 1

        # How many are we going to process ?
        if QueueProcessSize < len(self.data):
            ToDoCount = QueueProcessSize
        else:
            ToDoCount = len(self.data)

        try:
            if not self.HookStartProcessing(App, private_data, ToDoCount):
                return 0
        except:
            import sys, traceback
            info = sys.exc_info()
            (type, val, tb) = info
            zLOG.LOG('ZTxnQueue', zLOG.PANIC,
                "ZTxnQueue %s - HookStartProcessing Exception\n%s %s\n%s\n" %
                (self.path, type, val, 
                string.join(traceback.format_exception(type, val, tb), '\n')))
            get_transaction().abort()
            # Exception Initializing Process Function - Block the queue
            get_transaction().begin()
            self._recordException(info, PD=private_data)
            self.ProcessingState = 'B'
            self.HookException(App, private_data)
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()
            return 0

        (oid, data) = (None, None)
        Count = 0
        try:
            for idx in range(QueueProcessSize):
                rv = self.dequeue()
                if not rv:
                    break # Queue Empty
                (oid, data) = rv
                rvMessage = self.HookProcessOne(Data=data, OID=oid, private_data=private_data)
                if rvMessage != None:
                    # If anything is returned, just log it
                    zLOG.LOG('ZTxnQueue', zLOG.WARNING, str(rvMessage))
                Count = Count+1
            # Need to commit even if no items on the queue.
            # This releases the mutex
            get_transaction().commit()
            get_transaction().begin()
        except:
            import sys, traceback
            info = sys.exc_info()
            (type, val, tb) = info
            zLOG.LOG('PollingThread', zLOG.ERROR, "ZTxnQueue %s OID %s - Exception Occurred\n%s %s\n%s\n" %
                (self.path, oid, type, val,
                string.join(traceback.format_exception(type, val, tb), '\n')))

            # Wind the Tracked Dequeued back to the start in case of a conflict resolution
            get_transaction().abort()
            get_transaction().begin()
            self.ProcessingState = 'B'
            self._recordException(info, PD=private_data, OID=oid, Data=data)
            self.HookException(App, private_data)
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()
            return 0

        # Processing Queue Has completed - get the return value
        try:
            rv = self.HookEndProcessing(App, private_data, Count)
        except:
            import sys, traceback
            info = sys.exc_info()
            (type, val, tb) = info
            zLOG.LOG('ZTxnQueue', zLOG.PANIC,
                "ZTxnQueue %s - HookEndProcessing Exception\n%s %s\n%s\n" %
                (self.path, type, val, 
                string.join(traceback.format_exception(type, val, tb), '\n')))
            get_transaction().abort()
            # Exception Finalizing the Process - Block the Queue
            get_transaction().begin()
            self.ProcessingState = 'B'
            self._recordException(info, PD=private_data)
            self.HookException(App, private_data)
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()
            return 0
        return rv

class ZTxnQueue(ZTxnQueueBase):
    ### I am putting in these simpler functions to aid subclassing of the functionality.
    ### Rather than having to implement the full Process() functionality, you can
    ### just sub-class these.
    ### I am also thinking that these could be subclassed by a ZClass - How
    ### do you do that?

    # When Configuring the system, you can create five scripts in the
    # object to process the queue.
    #
    #   BeginBatch         - Called before processing a batch begins
    #   ProcessOne         - Process one item
    #   EndBatch           - Called when processing a batch is completed
    #   NotifyException    - Called if an exception occurs - not a handler
    #   StatusMessage      - Called to provide a status message

    security.declarePrivate('HookStartProcessing')
    def HookStartProcessing(self, App, private_data, Count):
        """Indicates whether processing should happen. Returns 0
        to break out of the loop, 1 to commence processing."""
        if not hasattr(self, 'BeginBatch'):
            return 1
        return self.BeginBatch(PD=private_data, Count=Count)

    security.declarePrivate('HookEndProcessing')
    def HookEndProcessing(self, App, private_data, Count):
        """Determines the return value for the Process method. It is
        called after the process method is completed, Transaction is
        also committed at this stage."""
        if not hasattr(self, 'EndBatch'):
            return len(self.data)
        return self.EndBatch(PD=private_data, Count = Count)

    security.declarePrivate('HookProcessOne')
    def HookProcessOne(self, Data, OID, private_data):
        """Override this to define your own processing logic."""
        if not hasattr(self, 'ProcessOne'):
            return len(self.data)
        return self.ProcessOne(Data=Data, OID=OID, PD=private_data)

    def getStatusMessage(self):
        """Return the status message for display in the TQPollingThread window"""
        if not hasattr(self, 'StatusMessage'):
            if self.ProcessingState == 'B':
                return "<b><font color=red>Blocked</font></b>&nbsp;&nbsp;Queue Size = %d" % len(self.data)
            else:
                return "Active&nbsp;&nbsp;Queue Size = %d" % len(self.data)
        return self.StatusMessage(ProcessingState = self.ProcessingState,
            QueueLength = len(self.data))

    security.declarePrivate('HookException')
    def HookException(self, App, private_data):
        """This method is called to if an exception occurs,
        which causes the queue to be blocked. It is intended for
        use in notifying an administrator."""
        if not hasattr(self, 'NotifyException'):
            return 0
        try:
            self.NotifyException(PD=private_data)
        except:
            zLOG.LOG('ZTxnQueue', zLOG.ERROR,
                "ZTxnQueue %s : ExceptionNotify Raised Exception - ignored" % self.path)

InitializeClass(ZTxnQueue)

manage_addZTxnQueueForm = PageTemplateFile('www/manage_addZTxnQueueForm.zpt', globals())

def manage_addZTxnQueue(dispatcher, id, title='', REQUEST=None):
    """Adds a new ZTxnQeueue object with id *id*.
    """
    id = str(id)
    ob = ZTxnQueue(id, title)
    dispatcher._setObject(id, ob)
    ob = dispatcher._getOb(id)
    # Register with the TQPollingThread
    ob._AmIManaged()
    if REQUEST is not None:
        return dispatcher.manage_main(dispatcher, REQUEST)

